// ForGet. A tool for managing development dependencies.
// Copyright (C) 2015  Daniel Hughes
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
using System;
using System.Xml.Linq;
using System.IO;
using System.Linq;
using System.Collections.Generic;

namespace ForGet
{
   public class Upgrader
   {
      public bool Upgrade(string configFile)
      {
         if(!File.Exists(configFile))
         {
            Console.Error.WriteLine("Error: Config file could not be found: {0}", configFile);
            return false;
         }
         XElement repo = XElement.Parse(File.ReadAllText(configFile));
         var repoPaths = ConfigHelpers.GetRepoPaths(repo);
         if(repoPaths.Any(r => r == null))
         {
            return false;
         }

         foreach(var dependency in repo.Elements("Dependency"))
         {
            if(!UpgradeDependency(dependency, repoPaths))
            {
               return false;
            }
         }
         File.WriteAllText(configFile, repo.ToString());
         return true;
      }

      public bool UpgradeDependency(XElement dependency, IEnumerable<string> repoPaths)
      {
         string id = dependency.Attribute("Id").Value;
         string versionString = dependency.Attribute("Version").Value;
         Version version = null;
         if(!Version.TryParse(versionString, out version))
         {
            Console.Error.WriteLine("Confile File contained invalid version {0} for dependency {1}", versionString, id);
            return false;
         }
         string upgradeType = dependency.Attribute("UpgradeType").Value;
       
         if(upgradeType != "Latest")
         {
            return true;
         }
       
         IEnumerable<Version> versions = 
            from filePath in GetDepCandidates(repoPaths, id)
            let versionFromRepo = GetVersion(id, filePath)
            where versionFromRepo != null
            orderby versionFromRepo descending
            select versionFromRepo;
         var latestRepoVersion = versions.FirstOrDefault();
         if(latestRepoVersion == null)
         {
            Console.WriteLine("{0} Failed to find package!", id, version);
            return false;
         }

         if(version == latestRepoVersion)
         {
            Console.WriteLine("{0} {1} Already at latest version", id, version);
            return true;
         }
         dependency.Attribute("Version").Value = latestRepoVersion.ToString();
         Console.WriteLine("{0} {1} Upgraded to version {2}", id, version, latestRepoVersion);
         return true;
      }
      
      IEnumerable<string> GetDepCandidates(IEnumerable<string> repos, string id)
      {
         return
            from repo in repos
            let repoPathforDependency = Path.Combine(repo, id)
            where Directory.Exists(repoPathforDependency)
            from filePath in Directory.GetFiles(repoPathforDependency)
            select filePath;
      }
    
      Version GetVersion(string id, string file)
      {
         string fileName = Path.GetFileNameWithoutExtension(file);
         string versionPart = fileName.Substring(id.Length + 1);
         Version version = null;
         if(!Version.TryParse(versionPart, out version))
         {
            return null; 
         }
         return version;
      }
   }
}

